body {
        background: white;
        margin: 2px;
        padding-left: 3em;
        padding-right: 3em;
}

h3 {
	clear: both;
	float: none;
}

#summary {
	font-size: 80%;
}

#summary h2 {
}

#results {
	float: left;
	clear: both;
}

.testsuite {
	margin-left: 1em;
	padding-left: 1em;
	margin-right: 1em;
	padding-right: 1em;
}

.testsuite-title {
	font-size: 115%;
	font-weight: bold;
	margin-top: 0.50em;
	margin-bottom: 0.25em;
	vertical-align: top;
	clear: both;
}

.testsuite-title table {
	vertical-align: top;
	width: 100%;
}

.testsuite-test-count {
	text-transform: none;
	text-align: right;
}

.testsuite-summary {
	width: 150px;
	text-align: right;
}

.testsuite-all-passed {
	
}

.test-case {
	margin-right: 1em;
	margin-left: 1em;
	clear: left;
	float: none;
}

.testsuite-some-failures {
	border-left-style: solid;
	border-right-style: solid;
	border-color: blue;
}

.testsuite-some-errors {
	border-left-style: solid;
	border-right-style: solid;
	border-color: red;
}

.test-error {
	float: right;
	clear: right;
	color: red;
}

.test-failure {
	float: right;
	clear: right;
	color: blue;
}

#problem-summary {
	float: left;
	clear: both;
}

#problem-summary body {
	float: none;
}

#failure-summary {
	margin-left: 1em;
	margin-right: 1em;
}

#failure-summary h3 {
	font-style: italic;
	font-weight: bold;
}

#error-summary {
	margin-left: 1em;
	margin-right: 1em;
}

#error-summary h3 {
	font-style: italic;
	font-weight: bold;
}

#footer {
	padding-top: 2em;
	text-align: right;
	font-size: medium;
	color: gray;
	clear: both;
}

.test-name {
	float: left;
	clear: none;
	display: block;
	width: 255px;
	overflow: hidden;
	height: 1.25em;
}

.test-time {
	float: left;
	padding-left: 20px;
	text-align: right;
	display: run-in;
	width: 35px;
}

.start-time {
	float: left;
	padding-left: 20px;
	padding-right: 20px;
	text-align: right;
	display: run-in;
//	width: 35px;
}


.test-space {
	padding-left: 20px;
	text-align: right;
	float: left;
	clear: none;
	display: block;
	width: 70px;
}